package mthsc865.datastructures;

import java.util.Collection;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.ConcurrentModificationException;

/**
 * This class implements a Stack using a dynamic array
 *
 * @author Daniel D. Warner
 * @version 2.0 09/07/04
 */
public class StackAL extends AbstractStack {
    /**
     * The dynamic array is implemented with the java.util.ArrayList
     */
    protected ArrayList v;
    
    /**
     * Constructs a stack whose dynamic array has a specified
     * initial size. Whenever the dynamic array is increased,
     * it's size will determined by the default ArrayList strategy,
     * which traditionally doubles the size.
     * @param capacity the initial size of the dynamic array.
     */
    public StackAL(int capacity) {
        hash = 0;
        v = new ArrayList(capacity);
    }
    
    /**
     * Constructs a stack whose dynamic array has the default
     * initial size for a ArrayList.
     */
    public StackAL() {
        hash = 0;
        v = new ArrayList();
    }
    
    /**
     * Constructs a stack which is initialized with the objects
     * in the specified collection
     * @param c the collection of objects to be initially pushed
     * onto this stack.
     */
    public StackAL(Collection c) {
        hash = 0;
        v = new ArrayList(c.size());
        Iterator ci = c.iterator();
        while (ci.hasNext()) {
            this.push( ci.next() );
        }
    }
    
    /**
     * Constructs a stack with a specified hash code and a clone
     * of the specified ArrayList.  This constuctor is only for use
     * within the class and its subclasses.
     * @param hash the hash code
     * @param v the ArrayList to be cloned.
     */
    protected StackAL(int hash, ArrayList v) {
        this.hash = hash;
        this.v = (ArrayList)v.clone();
    }
    
   /**
    * Pushes an object onto the top of the stack.
    * @param obj the object to be placed on the top of the stack.
    */
    public void push(Object obj) {
        updateHashCode(obj);
        v.add(obj);
    }
    
    /**
    * Removes and returns the object on the top of the stack.
    * @return the object on the top of the stack or null if the stack is empty.
    */
    public Object pop() {
        if (v.isEmpty()) {
            return null;
        } else {
            Object obj = v.remove(v.size()-1);
            downdateHashCode(obj);
            return obj;
        }
    }
    
    /**
     * Returns the object on the top of the stack.
     * The stack remains unchanged.
     * @return the object on the top of the stack or null if the stack is empty.
     */
    public Object peek() {
        if (v.isEmpty()) {
            return null;
        } else {
            return (v.get(v.size()-1));
        }
    }
    
    /**
     * Removes all the objects from this stack.
     * This is one of the optional operations specified by the Collection Interface
     */
    public void clear() {
        hash = 0;
        v.clear();
    }
    
    /**
     * The Cloneable Interface indicates that the clone
     * method, which is inherited from Object, is implemented
     * so that a true clone (a true and independent copy) of 
     * the object is returned.
     * @return a clone (a true and independent copy) of this stack.
     */
    public Object clone() {
        return (new StackAL(hash, v));
    }
    
    /**
     * An iterator for an ArrayList stack.
     */
    protected class StackALIterator implements Iterator {
        private StackAL myS;  // a reference to my Stack
        private int index;    // my current position in the stack 
        private int check;    // my copy of the initial hash code. It will be an
                              // error if the stack is altered during the iteration.
        /**
         * Constructor for an StackAL iterator
         * @param s the stack for this iterator.
         */
        protected StackALIterator(StackAL s) {
            myS = s;
            index = s.v.size();
            check = s.hash;
        }
        
        /**
         * Determines if there is an object in stack that can be returned by next().
         * @return true - if there is an object in the stack that can be returned by next().
         */
        public boolean hasNext() {
            //  Check that the stack has not been altered during the processing
            //  of this iterator
            if (check == myS.hash) {
                return (0 < index);
            } else {
                throw( new ConcurrentModificationException("This Stack has been altered"));
            }
        }
        
        /**
         * Returns the next object in stack, if there is one.
         * It throws a ConcurrentModificationException if the stack has been changed
         * during the iteration.
         * @return the next object in the stack.
         */
        public Object next() {
            if (check == myS.hash) {
                return myS.v.get(--index);
            } else {
                throw( new ConcurrentModificationException("This Stack has been altered"));
            }
        }
	
        /**
         * This method is an optional method of the Iterator interface.
         * If implemented, it would remove from the stack the last element returned
         * by the iterator. Since this would violate the integrity of the abstract stack
         * data structure it is not implemented.
         * If invoked this method throws an UnsupportedOperationException
         */
        public void remove() {
            throw( new UnsupportedOperationException(
            "remove is not implemented for mthsc865.datastructure.StackAL"));
        }
    }
    
    /**
     * A factory method which returns an Iterator to the collection in this stack.
     * This iterator will walk throught the stack from the top to the bottom.
     * It will throw a ConcurrentModificationException if the stack is altered during
     * the iteration.
     * @return a StackALIterator which will iterate through this stack starting from
     * the top.
     */
    public Iterator iterator() {
        return new StackALIterator(this);
    }
    
    /**
     * Determines the size of this stack.
     * @return the number of objects in this stack. 
     */
    public int size() {
        return v.size();
    }
    
    /**
     * List the objects in the stack
     * @return a formatted string listing the objects in this stack
     */
    public String toString() {
        String s = super.toString();
        return( "StackAL" + s.substring(5, s.length()) );
    }
    
    /**
     * This main method tests the StackAL class to insure that the
     * elementary functions are correct.
     * @param args optional command line arguments which will be ignored.
     */
    public static void main(String[] args) {
        StackAL s1 = new StackAL();
        String name;
        String linesep = System.getProperty("line.separator");
        
        // Place Alice, Bob, and Carol on the stack and print out the
        // stack and its basic properties.
        s1.push("Alice");
        s1.add("Bob");
        s1.push("Carol");
        System.out.println(s1);
        System.out.println("s1.hashCode: " + s1.hashCode());
        System.out.println("Empty? " + s1.isEmpty());
        
        // Test the iterator
        System.out.println(linesep + "s1 Iterator test:");
        Iterator iter = s1.iterator();
        while (iter.hasNext()) {
            name = (String)iter.next();
            System.out.println(name);
        }
        
        // Create a second and third stack
        System.out.println(linesep + "Looking at three stacks");
        StackAL s2 = new StackAL(s1);
        System.out.println(linesep + "s2 Iterator test:");
        iter = s2.iterator();
        while (iter.hasNext()) {
            name = (String)iter.next();
            System.out.println(name);
        }
        StackAL s3 = new StackAL();
        s3.addAll(s2);
        System.out.println(linesep + "s3 Iterator test:");
        iter = s3.iterator();
        while (iter.hasNext()) {
            name = (String)iter.next();
            System.out.println(name);
        }
        
        // Manipulate the stacks and check the different methods
        // particularly the hash code, equality, and contains.
        System.out.println("s1 equals s2? " + s1.equals(s2));
        System.out.println("s1 equals s3? " + s1.equals(s3));
        System.out.println("s1.hashCode: " + s1.hashCode());
        System.out.println("s2.hashCode: " + s2.hashCode());
        System.out.println("s3.hashCode: " + s3.hashCode());
        System.out.println("Is Bob in s1? " + s1.contains("Bob"));
        s1.add("Doug");
        System.out.println(s1.peek());
        System.out.println("s1.hashCode: " + s1.hashCode());
        System.out.println("s2.hashCode: " + s2.hashCode());
        System.out.println("s3.hashCode: " + s3.hashCode());
        System.out.println("s1 equals s2? " + s1.equals(s2));
        System.out.println("s1 equals s3? " + s1.equals(s3));
        System.out.println("Is s1 contained in s2? " + s2.containsAll(s1));
        System.out.println("Is s2 contained in s1? " + s1.containsAll(s2));
        System.out.println("Is s2 contained in s3? " + s3.containsAll(s3));
        
        // Check out the cloning mechanism
        System.out.println(linesep + "Looking at cloning");
        StackAL s4 = (StackAL)s1.clone();
        System.out.println("s1 equals s4? " + s1.equals(s4));
        System.out.println("s1.hashCode: " + s1.hashCode());
        System.out.println("s4.hashCode: " + s4.hashCode());
        System.out.println("s4.pop: " + s4.pop());
        System.out.println("s1 equals s4? " + s1.equals(s4));
        System.out.println("s1.hashCode: " + s1.hashCode());
        System.out.println("s4.hashCode: " + s4.hashCode());
        
        // Check out the conversion from a stack to an array.
        System.out.println(linesep + "Looking at an array and a stack");
        Object[] names = s1.toArray();
        for (int i=0; i<names.length; i++) {
            System.out.println(names[i]);
        }
        String[] moreNames = (String[])s1.toArray(new String[5]);
        System.out.println("moreNames.length: " + moreNames.length);
        for (int i=0; i<moreNames.length; i++) {
            System.out.println(moreNames[i]);
        }
        
        // Empty the stacks
        System.out.println(linesep + "Emptying the stacks");
        System.out.println(s1.pop());
        System.out.println(s1.hashCode());
        System.out.println(s1.pop());
        System.out.println(s1.pop());
        System.out.println(s1.pop());
        System.out.println("Empty? " + s1.isEmpty());
        s2.clear();
        System.out.println("Empty? " + s2.isEmpty());
        System.out.println("s1 equals s2? " + s1.equals(s2));
        System.out.println("s1.hashCode: " + s1.hashCode());
        System.out.println("s2.hashCode: " + s2.hashCode());
        
        // Test the iterator for consistency
        System.out.println(linesep + "Testing the iterator for consistency checking");
        iter = s4.iterator();
        try {
            System.out.println(iter.next());
            System.out.println(iter.next());
            System.out.println("Pop: " + s4.pop());
            System.out.println(iter.next());
        } catch (ConcurrentModificationException e) {
            System.out.println("Concurrent Modification Exception: " + e.getMessage());
        }
    }
}
