package mthsc865.datastructures;

import java.util.Collection;
import java.util.Iterator;
import java.lang.reflect.Array;

/**
 * Abstract class for the stack data structure.
 * It implements the push, pop and convenience methods.
 * The optional methods of the Collections interface, which would
 * allow the stack integrity to be violated are not implemented
 * and throw an UnsupportedOperation exception.
 *
 * @author Daniel D. Warner
 * @version 2.0 09/07/04
 */
abstract public class AbstractStack implements Cloneable, Collection {
    /**
     * The cached value of the hash code for the stack.  It is
     * assumed that push, pop, clear, and clone methods will
     * properly update this hash code.  The cached value of the
     * code should be used by iterators to check for any changes
     * in the stack while the iterator is being used.
     */
    protected int hash;  //  The cached hash code.
   
   /**
    * Pushes an object onto the top of the stack.
    * @param obj the object to be placed on the top of the stack.
    */
    abstract public void push(Object obj);
    
    /**
    * Removes and returns the object on the top of the stack.
    * @return the object on the top of the stack or null if the stack is empty.
    */
    abstract public Object pop();
    
    /**
     * Returns the object on the top of the stack.
     * The stack remains unchanged.
     *@return the object on the top of the stack or null if the stack is empty.
     */
    abstract public Object peek();
    
    /**
    * Pushes an object onto the top of the stack. Consistent with the 
    * Collection interface requirement that this method should ensure that
    * this collection contains the specified object.
    * @param obj the object to be placed on the top of the stack.
    * @return true because the stack has been changed.
     */
    public boolean add(Object obj) {
        this.push( obj );
        return true;
    }
    
    /**
    * Pushes each object of the specified collection onto the stack.
    * @param c the collection of objects to be placed on the stack.
    * @return true if the collection has any elements, because the 
    * stack has been changed.
    */
    public boolean addAll(Collection c) {
        Iterator ci = c.iterator();
        boolean changed = ci.hasNext();
        while (ci.hasNext()) {
            this.push( ci.next() );
        }
        return changed;
    }
    
    /**
     * Determines if the specified object is in the stack.
     * @param obj the object to be found in the stack.
     * @return true - if the specified object is in the stack.
     */
    public boolean contains(Object obj) {
        boolean b = false;
        Iterator iter = this.iterator();
        while (!b && iter.hasNext()) {
            b = obj.equals(iter.next());
        }
        return b;
    }
    
    /**
     * Determines if the specified collection is contained in the stack.
     * @param c the collection of objects to be found in the stack.
     * @return true - if the stack contains all of the objects in the
     * specified collection.
     */
    public boolean containsAll(Collection c) {
        Iterator ci = c.iterator();
        boolean b = true;
        while (b && ci.hasNext()) {
            b = this.contains(ci.next());
        }
        return b;
    }
    
    /**
     * The test for equality.
     * @param obj the object which may be equal to this stack.
     * @return true - if the specified object is equal to this stack.
     */
    public boolean equals(Object obj) {
        boolean b = (obj instanceof mthsc865.datastructures.AbstractStack) &&
            (this.size() == ((mthsc865.datastructures.AbstractStack)obj).size());
        if (b) {
            Iterator i = this.iterator();
            Iterator j = ((mthsc865.datastructures.AbstractStack)obj).iterator();
            while (b && i.hasNext()) {
                b = (i.next()).equals(j.next());
            }
        }
        return b;
    }
    
    /**
     * Returns the value of the hash code for this stack.
     * This method should return the same hash code for any two stacks in the
     * mthsc865.datastructures package that have the same objects stored in 
     * the same order.
     * @return hash - the hashcode for this stack
     */
    public int hashCode() {
        return hash;
    }
    
    /**
     * Updates the hash code for this stack
     * @param obj the object being added to the stack
     */
    protected void updateHashCode(Object obj) {
        long h = hash;
        hash = (int)((h + obj.hashCode()) % Integer.MAX_VALUE);
    }
    
    /**
     * Downdates the hash code for this stack
     * @param obj the object being removeed from the stack
     */
    protected void downdateHashCode(Object obj) {
        long h = hash;
        hash = (int)((h - obj.hashCode()) % Integer.MAX_VALUE);
    }	
    
    /**
     * Determines whether the stack is empty.
     * @return true - if this collection contains no elements.
     */
    public boolean isEmpty() {
        return (this.size() == 0);
    }
    
    /**
     * This is an optional method of the Collection Interface.
     * If implemented, it would remove a single instance of the specified object
     * from the stack. Since this would violate the integrity of the abstract stack 
     * data structure, it is not implemented.
     * @param obj the object to be removed.
     * @return throws an UnsupportedOperationException
     */
    public boolean remove(Object obj) {
        throw( new UnsupportedOperationException(
            "remove is not implemented for mthsc865.datastructure.AbstractStack"));
    }
    
    /**
     * This is an optional method of the Collection Interface.
     * If implemented, it would remove a single instance of each item in the specified
     * collection from the stack. Since this would violate the integrity of the abstract
     * stack data structure, it is not implemented.
     * @param c the collection of objects to be removed.
     * @return throws an UnsupportedOperationException
     */
    public boolean removeAll(Collection c) {
        throw( new UnsupportedOperationException(
            "removeAll is not implemented for mthsc865.datastructure.AbstractStack"));
    }
    
    /**
     * This is an optional method of the Collection Interface.
     * If implemented, it would remove all of the objects in the stack except for the
     * objects in the specified collection. Since this would violate the integrity of
     * the abstract stack data structure, it is not implemented.
     * @param c the collection of objects to be retained.
     * @return throws an UnsupportedOperationException
     */
    public boolean retainAll(Collection c) {
        throw( new UnsupportedOperationException(
            "retainAll is not implemented for mthsc865.datastructure.AbstractStack"));
    }
    
    /**
     * Creates an array containing the objects in this stack.
     * @return an array of objects.
     */
    public Object[] toArray() {
        Object[] newArray = new Object[this.size()];
        Iterator iter = this.iterator();
        int k=0;
        while (iter.hasNext()) {
            newArray[k++] = iter.next();
        }
        return newArray;
    }
    
    /**
     * Creates an array containing the objects in this stack.
     * @param a an array of objects whose run type is compatible with all the
     * objects in the stack. 
     * @return an array of objects of the same type as the elements of a.
     * If a is large enough, then all the objects in the stack are copied into
     * a, with nulls stored in any extra empty entries of a.  If a is not large
     * enough, then the method returns a new array whose objects have the same 
     * type as a.
     */
    public Object[] toArray(Object[] a) {
        Iterator iter = this.iterator();
        int k = 0;
        if (this.size() <= a.length) {
            while (iter.hasNext()) {
                a[k++] = iter.next();
            }
            while (k < a.length) {
                a[k++] = null;
            }
            return a;
        } else {
            Class c = a.getClass();
            c = c.getComponentType();
            Object[] newArray = (Object[])Array.newInstance(c, this.size());
            while (iter.hasNext()) {
                newArray[k++] = iter.next();
            }
            return newArray;
        }
    }
    
    /**
     * List the objects in the stack
     * @return a formatted string listing the objects in this stack
     */
    public String toString() {
        String s = "Stack {";
        String lineSep = System.getProperty("line.separator");
        Iterator i = this.iterator();
        while (i.hasNext()) {
            s += lineSep + "	" + i.next().toString();
        }
        return (s + " }");
    }	
}
